function [D,X,err] = RotateSVD(params)
% This function implments the R-SVD (Rotate-SVD) method in
%  G. Grossi, R. Lanzarotti, J. Lin. Orthogonal Procrustes analysis for
%  dictionary learning in sparse representation, manuscript under review
% that uses the Orthogonal Procrustes shape analysis in the alternating
% optimization scheme for dictionary learning in sparse models.
%
% Syntax:
%    [D,X,err] = RotateSVD(params)
%    ARGUMENTS
%        params is a struct with mandatory fields:
%          data:     training set matrix (Y) with examples in columns
%          Tdata:    sparsity bound (k) for sparse representation of each
%                    example
%          dictsize: number (m) of atoms in the dictionary D
%        and optional fields
%          iternum:     number (T) of iterations in the R-SVD's
%                       alternating. Default = 100
%          SubdictSize: size (s) of each sub-dictionary in Step 2 of
%                       R-SVD. Default = 10
%          sparsifier:  sparse decomposition algorithm name, either 'omp'
%                       (default) for Orthogonal Matching Pursuit or
%                       'klimaps' for k-LiMapS (see Adamo et al.)
%          max_iter:    number of iterations in the k-LiMapS sparse
%                       decomposition algorithm. Default = 100
%    RETURNED VARIABLES
%        D:   dictionary learnt with R-SVD from the training set Y
%        X:   the i-th column of X is the sparse encoding of the i-th
%             example in Y
%        err: iternum-length vector with RMSE at each iteration
% REQUIRES:
%    rsvd()      Truncated SVD by random structured matrices (code by
%                Antoine Liutkus). Algorithm described in:
%                Halko, N., Martinsson, P.-G., Tropp, J. A., 2011. Finding
%                structure with randomness: Probabilistic algorithms for
%                constructing approximate matrix decompositions. SIAM Rev.
%                53 (2), 217-288
%    OMP-Box v10 Implementation of the Batch-OMP and OMP-Cholesky
%                algorithms for quick sparse-coding of large sets of
%                signals. (Necessary for 'omp' sparsifier)
%    klimaps     A fixed-point based iterative scheme for k-sparse
%                decomposition. (Necessary for 'klimaps' sparsifier)
% Jianyi Lin, University of Milan, Italy (jianyi.lin@unimi.it)
% Dec 30, 2015

%% Parameters retrieval
Y = params.data;           % Training set as columns of matrix Y
k = params.Tdata;          % Upper bound on sparsity level
m = params.dictsize;       % Dictionary size
if isfield(params,'iternum')
    nIter = params.iternum;    % Number of iterations (alternating steps)
else
    nIter = 100;
end
if isfield(params,'SubdictSize')
    s = params.SubdictSize;    % Subdictionary size parameter (# atoms to be taken at each rotation)
else
    s = 10;
end
if isfield(params,'sparsifier')
    sparsifier = params.sparsifier;
else
    sparsifier = 'omp';
end
if isfield(params,'max_iter')
    klimapsIter = params.max_iter;
else
    klimapsIter = 100;
end

G = ceil(m/s); %number of subdictionaries
%[n,L] = size(Y);
err = zeros(1,nIter);

% Parameters for OMP
if strcmp(sparsifier,'omp')
    ompparams = {'checkdict','off'};
    msgdelta = -1;
    ompparams{end+1} = 'messages';
    ompparams{end+1} = msgdelta;
    ompfunc = @omp;
end

% Initial dictionary formed by randomly picking examples in Y
data_ids = find(sum(Y.^2,1) > 1e-6);   % ensure no zero-column is chosen
perm = randperm(length(data_ids));
D = normcols(Y(:,data_ids(perm(1:m))));

for i = 1:nIter
    %% Step 1. Sparse Coding
    switch sparsifier
        case 'omp'
            Gram = D'*D;
            %X = ompfunc(D,Y,Gram,k,ompparams{:});
            X = ompfunc(D'*Y,Gram,k,ompparams{:});
        case 'klimaps'
            X = klimaps(D,pinv(D),Y,k,klimapsIter);
        otherwise
            error(['Wrong sparsifier ' sparsifier ' specified.']);
    end
    
    %% Step 2. Dictionary Update
    % Ordering/selecting atoms to be updated by
    %  ascending popularity (column) vector
    Pop = sum(abs(X)>10^-6,2); % exclude numerical zeros
    [~, M] = sort(Pop,'ascend');
    sizeM = numel(M);
    
    % Determine block limits
    blockSizes = [ones(1,G-1)*s, sizeM-(G-1)*s];
    blockEnds = cumsum(blockSizes);
    blockStarts = [0 blockEnds(1:end-1)] + 1;
    
    % Pre-computations
    XYt = X*Y';
    XXt = X*X'; % Gram matrix
    
    for g = 1:G
        I = M(blockStarts(g):blockEnds(g));
        IC = true(1,m);
        IC(I) = false;
        I = not(IC);
        
        % Solve the orthogonal Procrustes problem: argmin ||Y' - X'*Q||_F^2 s.t. Q'*Q=I
        HEt = D(:,I)*(XYt(I,:)-XXt(I,IC)*D(:,IC)'); % Efficient matrix products
        
        % Full SVD decomposition
        %[U,~,V] = svd(HEt);
        %V = V(:,1:s);
        %U = U(:,1:s);
        % Compact SVD decomposition (up to the rank<=s)
        %[U,~,V] = svds(HEt,s); %inefficient
        [U,~,V] = rsvd(HEt,sum(I)); % efficient Truncated SVD by structured random matrices
        
        D(:,I) = V*U'*D(:,I); % standard Procrustes solution
    end%for
    
    %compute errors
    %err(i) = norm(Y-D*X,'fro')/sqrt(numel(Y));
    err(i) = compute_err(D,X,Y);
end%for

end%function

%% Auxiliary functions for error evaluation from K-SVD by
%%  Ron Rubinstein
%%  Computer Science Department
%%  Technion, Haifa 32000 Israel
%%  ronrubin@cs 
%% replicated here
function err = compute_err(D,X,Y)
    % Compute the RMSE: square-root mean of Frobenius norm =
    % ||Y-D*X||_F / sqrt(numel(Y))
    err = sqrt(sum(reperror2(Y,D,X))/numel(Y));
end

function err2 = reperror2(Y,D,X)
    % Compute the squared-2-norm for every example
    % Compute in column-blocks to conserve memory
    err2 = zeros(1,size(Y,2));
    blocksize = 2000;
    for i = 1:blocksize:size(Y,2)
        blockids = i : min(i+blocksize-1,size(Y,2));
        err2(blockids) = sum((Y(:,blockids) - D*X(:,blockids)).^2);
    end
end

function y = normcols(x)
    %NORMCOLS Normalize matrix columns.
    %  Y = NORMCOLS(X) normalizes the columns of X to unit length, returning
    %  the result as Y.
    y = x*spdiag(1./sqrt(sum(x.*x)));
end

function Y = spdiag(V,K)
    %SPDIAG Sparse diagonal matrices.
    %   SPDIAG(V,K) when V is a vector with N components is a sparse square
    %   matrix of order N+ABS(K) with the elements of V on the K-th diagonal. 
    %   K = 0 is the main diagonal, K > 0 is above the main diagonal and K < 0
    %   is below the main diagonal. 
    %
    %   SPDIAG(V) is the same as SPDIAG(V,0) and puts V on the main diagonal.
    if (nargin<2)
      K = 0;
    end

    n = length(V) + abs(K);

    if (K>0)
      i = 1:length(V);
      j = K+1:n;
    elseif (K<0)
      i = -K+1:n;
      j = 1:length(V);
    else
      i = 1:n;
      j = 1:n;
    end

    Y = sparse(i,j,V(:),n,n);
end